use std::string::ToString;

use super::BaseType;
use super::Quantifier;
use super::Expr;
use super::Variable;
use super::GazType;

pub enum Stat {
    Block(Box<Block>),
    Declaration(Box<Declaration>),
    TypeDef(Box<TypeDef>),
}

impl Stat {
    pub fn new_block(x: Block) -> Self {
        Self::Block(Box::new(x))
    }

    pub fn new_decl(x: Declaration) -> Self {
        Self::Declaration(Box::new(x))
    }

    pub fn new_typedef(x: TypeDef) -> Self {
        Self::TypeDef(Box::new(x))
    }
}

impl ToString for Stat {
    fn to_string(&self) -> String {
        match self {
            Stat::Block(x) => x.to_string(),
            Stat::Declaration(x) => x.to_string(),
            Stat::TypeDef(x) => x.to_string(),
        }
    }
}

#[derive(Default)]
pub struct GlobalBlock {
    statements: Vec<Stat>,
}

impl GlobalBlock {
    pub fn push(&mut self, node: Stat) {
        self.statements.push(node);
    }
}

impl ToString for GlobalBlock {
    fn to_string(&self) -> String {
        let mut s = String::new();

        for stat in &self.statements {
            s.push_str(&stat.to_string());
        }
        s
    }
}

pub struct Block {
    statements: Vec<Stat>,
}

impl Block {
    pub fn push(&mut self, node: Stat) {
        self.statements.push(node);
    }
}

impl ToString for Block {
    fn to_string(&self) -> String {
        let mut s = String::from("{ ");

        for stat in &self.statements {
            s.push_str(&stat.to_string());
            s.push(' ');
        }

        s.push('}');
        s
    }
}

#[derive(Default, Builder)]
pub struct Declaration {
    type_ident: String,
    variable: Variable,
    assn: Option<Expr>,
}

impl ToString for Declaration {
    fn to_string(&self) -> String {
        let mut s = String::new();

        s.push_str(&self.variable.get_quantifier().to_string());
        s.push(' ');
        s.push_str(&self.type_ident);
        s.push(' ');
        s.push_str(&self.variable.to_string());

        if self.assn.is_some() {
            s.push_str(" = ");
            s.push_str(&self.assn.as_ref().unwrap().to_string());
        }

        s.push(';');
        s
    }
}

#[derive(Clone, Default, Builder)]
pub struct TypeDef {
    old_name: String,
    new_name: String,
    basetype: BaseType,
}

impl ToString for TypeDef {
    fn to_string(&self) -> String {
        let mut s = String::from("typedef ");
        s.push_str(&self.old_name);
        s.push(' ');
        s.push_str(&self.new_name);
        s.push(';');
        s
    }
}
