use super::*;

/// Generates values from regions of interest for an integer
///
/// These are concentrated around i32::MIN, 0, and i32::MAX, though any i32 value is possible
pub struct IntGenerator {
    distro_pos: rand_distr::Normal<f32>,
    distro_zero: rand_distr::Normal<f32>,
    distro_neg: rand_distr::Normal<f32>,
}

impl IntGenerator {
    pub fn new(p: &Params) -> Self {
        Self {
            distro_pos: rand_distr::Normal::new(0.0, p.dev.int_gen_distro_pos_stddiv).unwrap(),
            distro_zero: rand_distr::Normal::new(0.0, 3.0).unwrap(),
            distro_neg: rand_distr::Normal::new(0.0, p.dev.int_gen_distro_neg_stddiv).unwrap(),
        }
    }
}

impl Distribution<i32> for IntGenerator {
    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> i32 {
        let p: f64 = rng.gen();

        if p < 0.3 {
            i32::MAX - (self.distro_pos.sample(rng).abs() as i32)
        } else if p < 0.7 {
            self.distro_zero.sample(rng) as i32
        } else {
            -1 * (i32::MAX - (self.distro_pos.sample(rng).abs() as i32))
        }
    }
}

/// Generates values from regions of interest for a float
///
/// These are concentrated around f32::MIN, 0, and f32::MAX, though any f32 value is possible
pub struct FloatGenerator {
    distro_pos: rand_distr::Normal<f32>,
    distro_zero: rand_distr::Normal<f32>,
    distro_neg: rand_distr::Normal<f32>,
}

impl FloatGenerator {
    pub fn new(p: &Params) -> Self {
        Self {
            distro_pos: rand_distr::Normal::new(1.0, p.dev.float_gen_distro_pos_stddiv).unwrap(),
            distro_zero: rand_distr::Normal::new(0.0, 3.0).unwrap(),
            distro_neg: rand_distr::Normal::new(1.0, p.dev.float_gen_distro_neg_stddiv).unwrap(),
        }
    }
}

impl Distribution<f32> for FloatGenerator {
    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> f32 {
        let p: f64 = rng.gen();

        if p < 0.33 {
            f32::MAX - self.distro_pos.sample(rng).abs()
        } else if p < 0.66 {
            self.distro_zero.sample(rng)
        } else {
            -1.0 * (f32::MAX - self.distro_neg.sample(rng).abs())
        }
    }
}
